# Copyright (c) 2001-2003, RedIRIS  (original PAPI module)
# Copyright (c) 2004, Telefonica Moviles Espaa  (this module)
#
# You may distribute under the terms of either the GNU General Public License
# or the Artistic License, as specified in the LICENSE file that was shipped
# with this distribution.
#
# ---------------------------------------------------------------------------------
#
# This module is based on PAPI::RewritingProxy (http://papi.rediris.es)
# (original from Rodrigo Castro and Diego R. Lopez).
#
# * Changelog (version 0.1) [03.Jun.2004]
#   =====================================
# 
# - Configuration is read from PerlSetVar/PerlAddVar directives instead of PoA object.
#   No need of other PAPI modules.
# - Name of authentication cookie is placed in $TMECookie. Default: "Apache::AuthCookieLDAP_APACHE".
# - HTTPS support for Remote_Domain (new directive: "Remote_Domain_HTTPS").
# - Rewrite now works for HTTPS as well as HTTP.
# - New directive "Raw_Redirect_All" in order to apply rewrite directly to the
#   contents of web page (instead of using Tokeparser library).
# - New directive "Debug" to make debugging of this module easier.
#
# * Changelog (version 0.2): [10.Jun.2004]
#   ======================================
#
# - Fixed empty/missing Content-type header.
# - Fix to avoid underscore to dash conversion in HTTP headers' fields.
# - Case insensitive when comparing HTTP headers' fields (eg: /^Cookie$/i) (web clients
#   sometimes send "cookie:" instead of "Cookie:" -> cookie substitution was broken).
# - Removed "eval" from HTTP_Auth directive.
# - More "debug" lines.
# - All comments translated into English.

# * Changelog (version 0.3): [08.Jul.2004]
#   ======================================
#
# - Ready for libwww-perl 5.80 (HTTP::Headers fix).
# - When using auto-form, all possible cookies are sent in one and only Cookie header.
# - New directive "Post_Redirect" (PerlAddVar), equivalent to PAPI_Redirect but applied to POST info being sent.
# - Content-Length fix (this field was sent twice in a POST request).
#
# * Changelog (version 0.4): [04.Nov.2004]
#   ======================================
#
# - Fixed repeated headers ("Host" and "User-Agent" were being sent twice).
# - New possible value for "Debug" directive (2 == print request and response contents).
# - New directive "Cookie_AddString". It adds a given string to all cookies (eg: SESSION_ID_10234) exchanged
#   between the proxy and the browser in order to differentiate cookies across similar services offered at
#   different vhosts. Cookies are cleaned and come back to their original states before being sent to the final
#   application so it should be a transparent mechanism.
#
# * Changelog (version 0.5): [01.Dic.2004]
#   ======================================
#
# - Fix: replace for "Remote_URL" string was only in effect for HTML content (so it was needed an additional
#   PAPI_Redirect directive for non-HTML content or when Raw_Redirect_All was active). Now it works as expected.
# - Fix: code for empty/missing Content-type header check was buggy. Now only applies to code 200 responses.
# - New directive "NoCache". When active (i.e, value 1), the reverse proxy will include "Pragma: no-cache" and 
#   "Cache-Control: no-cache" headers in requests sent to the final server, asking to avoid response caching.
#   May be useful when debugging.
# - Only print parsed response when dealing with HTML content, content-type specified by Rewrite_MIME_Types or
#   if the url matches Rewrite_URL_Pattern.
#
# * Changelog (version 0.6): [20.Dic.2004]
#   ======================================
#
# - Added comments to source code. Clean-up.
# - Documentation has been reviewed (added configuration directives, etc).
# - More robust input (removing trailing "/" when necessary, etc).
# - Fix: Rewrite_MIME_Types only worked when one MIME-type was entered. Now it accepts several MIME-type strings,
#   separated by whitespace.
# - Fix: URL translation was not correctly working when Raw_Redirect_All = 1 and Strip_Location = 1.
# - New directives: "PAPI_Redirect_Delim", "PAPI_Redirect_Extended", "Post_Redirect_Delim", "Post_Redirect_Extended",
#   "Rewrite_MIME_Types_Extended" and "Rewrite_URL_Pattern_Extended".
# - **IMPORTANT**: By default, "Extended mode" is OFF. In past versions, all is taken as if extended mode was in use
#   so behaviour has been changed. Be careful!
#
# * Changelog (version 0.7): [16.Sep.2005]
#   ======================================
#
# - Fix: HTTPS requests through proxy (CONNECT method) didn't work.
# - Fix: Proxy authentication didn't work (nor HTTP neither HTTPS).
# - New directives: "Proxy_User" and "Proxy_Pass". Used to perform proxy authentication.
#
# 
# Author: Roman Medina-Heigl Hernandez (Telefonica Moviles Espaa).
#

package Apache::RewritingProxyTME;

use strict;
use Apache::Constants qw(:common);
use Apache::Util qw(:all);
use Apache::Log ();
use Apache::URI;
use LWP::UserAgent;
use URI::URL;
use HTML::TokeParser;
use HTML::TreeBuilder;
use HTTP::Cookies;
use HTTP::Request::Common;
use HTTP::Request::Form;
use CGI;
use vars '$host_dest';                                  # 'http://host.domain:port'   ("server URL" to fetch by proxy, without URI)
use vars '$path_dest';                                  # '/'   (URI in "proxy URL", i.e, where to map the "/" of final server)
use vars '$name_dest';                                  # 'host'   (the server name without the domain part)
use vars '$debug';                                      # Level of debugging: 0 [none] (default), 1 [normal] or 2 [full]
use vars '$old_LWP';                                    # 1 if using libwww-perl <= 5.79, 0 otherwise
use vars '$secure';                                     # 1 if $host_dest ("server URL") is "https" (secure access)
use vars qw($VERSION @ISA @EXPORT);
$|=1;

$Apache::RewritingProxyTME::VERSION = '0.7';

sub handler {
  my $r = shift;                                        # Let's suppose we are requesting 'http://proxy/portal/index.html'

  $debug = $r->dir_config('Debug');

  my $urlToFetch = $r->uri;                             # $urlToFetch = '/portal/index.html'
  if ($r->dir_config('Remote_URL')) {
    $host_dest = $r->dir_config('Remote_URL');          # E.g. 'http://1.2.3.4:9090'
    my @thn = ($host_dest=~m(^(.*?)://(.*?)\.(.*?)$));  # @thn = ['http'  '1'  '2.3.4:9090']
    $name_dest = $thn[1];                               # $name_dest = '1'
    $secure = (lc($thn[0]) eq 'https' ? 1 : 0);
    ### The final purpose is to access 'http://1.2.3.4:9090/portal/index.html'
  }
  elsif ($r->dir_config('Remote_Domain')) {             # E.g. 'example.com:8080'
    my @thn = ($urlToFetch=~m(([^/]+)/(.*?)$));         # @thn = ['portal'  'index.html']
    $host_dest = ($r->dir_config('Remote_Domain_HTTPS')?'https':'http') . "://" . $thn[0] . "." . $r->dir_config('Remote_Domain');
    # $host_dest = 'http://portal.example.com:8080'
    $name_dest = $thn[0];                               # $name_dest = 'portal'
    $urlToFetch = "/" . $thn[1];                        # $urlToFetch = '/index.html'
    $secure = ($r->dir_config('Remote_Domain_HTTPS') ? 1 : 0 );
    ### The final purpose is to access 'http://portal.example.com:8080/index.html'
  }

  $path_dest=$r->location;                              # $path_dest = '/' (comes from <Location /> tag in httpd.conf [vhost conf])

  # Let's sanitize our input to avoid missing or repeated '/'
  # First, we strip $host_dest so it doesn't contain one or more trailing '/'
  $host_dest =~ s%/+$%%g;

  # Location must end with one only '/'. Better to force it...
  $path_dest =~ s%/+$%%g;
  $path_dest .= '/';

  $urlToFetch =~ s/$path_dest//;
  return &fetchURL($r, $urlToFetch);			# $urlToFetch is the URI without the first '/'
}



###########################
# sub escape_string
#
# Parameters:
#
# $string - string to escape
#
# Returns:
#
# $string - processed string

sub escape_string
{
  my $string = shift;

  $string =~ s/([\[\]\.\?\*\+\^\$\|\-\&\#\{\}\(\)\\])/\\$1/g;
  return $string;
}



###########################
# sub rewriteString
#
# Parameters:
#
# $r - an Apache request object
# $directive - Name of directive
# $outString - string to process
#
# Returns:
#
# $outString (processed string)

sub rewriteString {
  my $r = shift;
  my $directive = shift;
  my $outString = shift;
  my $delim;

  if ($r->dir_config($directive.'_Delim')) {            # PAPI_Redirect_Delim, Post_Redirect_Delim
    $delim = $r->dir_config($directive.'_Delim');
  } else {
    $delim = '\s+';                                     # By default, one or more white-space chars are used as the delimiter string
  }

  if ( $r->dir_config($directive.'_Extended') && $r->dir_config($directive.'_Extended') eq "1" ) {
    # Extended mode. Use of special strings is permitted (\s, \S, \w, \W, $1, ...)
    for my $redirect ($r->dir_config->get($directive)) {
      my ($rde, $rdv) = ($redirect =~ /(\S+)$delim(.*)/);
      my $evs = "\$outString =~ s($rde)($rdv)gsx";
      my $num_matches = eval $evs;
      $r->log->notice("(Extended) Rewrite ($directive $rde $rdv): $num_matches " . ($num_matches == 1 ? 'match': 'matches')) if ($debug and $num_matches);
    }
  } else {
    # Normal mode (default). All characters are processed "as-is". No special strings are meaningful
    # (for instance, "\s" is a literal "\" followed by the "s" char; it is not a white-space char)
    for my $redirect ($r->dir_config->get($directive)) {
      my ($orig, $repl) = split qr($delim), $redirect, 2;
      my $orig_escaped = escape_string($orig);
      my $num_matches = ($outString =~ s/$orig_escaped/$repl/g);
      $r->log->notice("(Normal) Rewrite ($directive $orig $repl): $num_matches " . ($num_matches == 1 ? 'match': 'matches')) if ($debug and $num_matches);
    }
  }

  return $outString;
}



##################################
# sub fetchURL
#
# Parameters:
#
# $r - an Apache request object
# $url - initially the URI to fetch without the initial '/'. Afterwards, the URL to fetch and process
#
# Returns:
#
# OK if it's happy.
# an appropriate code if other than 200 is received.

sub fetchURL {
  my $r = shift;
  my $url = shift;                                      # E.g. $url = 'portal/index.html'
  my $ua = new LWP::UserAgent;

  if ($debug) {
    $r->log->notice("Entering sub fetchURL [ $url ]");
    $r->log->notice("host_dest = $host_dest");
    $r->log->notice("path_dest = $path_dest");
    $r->log->notice("name_dest = $name_dest");
  }

  # Authentication cookie
  my $TMECookie = 'Apache::AuthCookieLDAP_APACHE';
 
  # Let's make sure we send all redirects back to the browser
  $ua->requests_redirectable([]);

  # Let's carry the User Agent to the server also.
  my $browser = $r->header_in('User-Agent');
  $ua->agent($browser);

  # Use HTTP authentication if defined
  my $tauri = URI::URL->new($host_dest);
  for my $hauth ($r->dir_config->get('HTTP_Auth')) {            # E.g. HTTP_Auth = 'realm::username::password'
    my ($br,$bu,$bp) = split(/\s*::\s*/,$hauth);                # It will also accept white-spaces ('realm :: username  ::password')
    $r->log->notice("HTTP Auth: $br // $bu // $bp") if $debug;
    $ua->credentials($tauri->host_port(),$br,$bu,$bp);
  }

  # We have to append the query string since it got munged by apache when this was first requested.
  my $rurl = $url;

  # Avoid double slashes into URL (seems to have problems in Windows boxes)
  $rurl =~ s%/{2,}%/%g;
  $rurl .= "?". $r->args if ($r->args && $url !~ /\?/);

  if ($r->dir_config('Strip_Location') && $r->dir_config('Strip_Location') eq "1") {
    $url="$host_dest/$rurl";
  } else {
    $url="$host_dest$path_dest$rurl";
  }
  
  my ($up1, $hostname) = split (/\/\//, $host_dest);

  my($res,$req);
  $r->log->notice("-*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*-") if $debug;
  $r->log->notice("Fetching remote URL: $url");
  $r->log->notice("-*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*---*-") if $debug;
  
  # Fix to avoid underscore to dash conversion and other oddities in HTTP headers' fields.
  if ($HTTP::Headers::VERSION lt '1.61') {
    # libwww-perl <= 5.79
    $old_LWP = 1;
    $HTTP::Headers::TRANSLATE_UNDERSCORE = 0;
  } else {
    # libwww-perl >= 5.80
    $old_LWP = 0;
  }
  $req = HTTP::Request->new($r->method, $url);

  # Use a proxy server if defined
  if ($r->dir_config('Proxy_Server') && $r->dir_config('Proxy_Server') ne "") {
    if ($secure) {
      # HTTPS request
      $ENV{HTTPS_PROXY} = $r->dir_config('Proxy_Server');           # E.g. Proxy_Server = 'http://proxy:8080/'
      $ENV{HTTPS_PROXY_USERNAME} = $r->dir_config('Proxy_User');
      $ENV{HTTPS_PROXY_PASSWORD} = $r->dir_config('Proxy_Pass');
      $r->log->notice("Sending HTTPS request through proxy [".$r->dir_config('Proxy_Server')."]") if $debug;
    } else {
      # HTTP request
      $ua->proxy(['http'] => $r->dir_config('Proxy_Server'));       # E.g. Proxy_Server = 'http://proxy:8080/'
      $req->proxy_authorization_basic( $r->dir_config('Proxy_User'), $r->dir_config('Proxy_Pass') );
      $r->log->notice("Sending HTTP request through proxy [".$r->dir_config('Proxy_Server')."]") if $debug;
    }
  }

  # Process and tidy headers:
  # Do not pass to the remote server TMECookie
  # Block "Host:" and "User-Agent:" since they are automagically written by LWP::UserAgent
  # Block all "Connection:" headers
  # Since we need to parse the HTML response, block "Accept-Encoding", so
  # we don't have to uncompress data before parsing it
  my %hdrs = $r->headers_in();
  $r->log->notice("***** Parsing in-headers *****") if $debug;
  for my $hdr (keys %hdrs) {
    $r->log->notice("Hdr-in: $hdr // Val: " . $hdrs{$hdr}) if $debug;
    if ($hdr =~ /^Cookie$/i) {
      my $fcv = "";
      for my $ckv (split /[\s]*;[\s]*/,$hdrs{$hdr}) {
        my ($cn, $cv) = split /[\s]*=[\s]*/,$ckv,2;
        if ($cn !~ /$TMECookie/i) {
          $r->log->notice("Passed: cn=$cn // cv=$cv") if $debug;
          if ($r->dir_config('Cookie_AddString')) {
            # Remove suffix string
            my $suffix = $r->dir_config('Cookie_AddString');
            $cn =~ s/_$suffix$//;
            $r->log->notice("Removed suffix: cn=$cn // cv=$cv") if $debug;
          }
          if ($fcv ne "") {
            $fcv .= "; $cn=$cv";
          } else {
            $fcv = "$cn=$cv";
          }
        }
        else {
          $r->log->notice("Blocked: cn=$cn // cv=$cv") if $debug;
        }
      }
      $req->header ($old_LWP ? $hdr : ':'.$hdr, $fcv) if ($fcv ne "");
      $r->log->notice("Hdr-out: $hdr // Val: " . $fcv) if ($debug and $fcv ne "");
    }
    elsif ( $hdr !~ /^Host$/i && $hdr !~ /^User-Agent$/i && $hdr !~ /^Connection$/i && $hdr !~ /^Accept-Encoding$/i && $hdr !~ /^Content-Length$/i && ( ($hdr !~ /^Pragma$/i && $hdr !~ /^Cache-Control$/i ) || !$r->dir_config('NoCache')) ) {
      $req->header ($old_LWP ? $hdr : ':'.$hdr, $hdrs{$hdr});
      $r->log->notice("Hdr-out: $hdr // Val: " . $hdrs{$hdr}) if $debug;
    }
  }

  if ($r->dir_config('NoCache')) {
    for my $hdr ('Pragma', 'Cache-Control') {
      $req->header ($old_LWP ? $hdr : ':'.$hdr, 'no-cache');
      $r->log->notice("Hdr-out [forced]: $hdr // Val: " . 'no-cache') if $debug;
    }
  }

  # In case of POST request we need to rewrite data according to "Post_Redirect" directive
  if ($r->method eq 'POST'){
    my $len = $r->header_in('Content-Length');
    my $buf;
    $r->read($buf,$len);
    $req->content(&rewriteString($r, 'Post_Redirect', $buf));
  }

  # Now we launch the request to web server
  $res = $ua->request($req);

  if ($debug && $debug == 2) {
    $r->log->notice('***** Request sent to server *****');
    $r->log->notice("=======================================================================");
    for my $line (split(/\n/, $req->as_string)) {
      $r->log->notice($line);
    }
    $r->log->notice('[Host and User-Agent headers are not showed although they exist]');
    $r->log->notice("=======================================================================");
  }

  # Parse "Form_Processor" directives
  my $parse_fpd = {};
  for my $formprocessor_line ($r->dir_config->get('Form_Processor')) {
    my ($parse_urldef,$parse_val) = split /\s+/,$formprocessor_line,2;
    my ($parse_tfn,$parse_url2) = split /@/,$parse_urldef;
    if (defined $parse_url2) { $parse_urldef = $parse_url2; }
    else { $parse_urldef = $parse_tfn; $parse_tfn = undef; }
    $parse_fpd->{$parse_urldef}{formName} = $parse_tfn;
    $parse_fpd->{$parse_urldef}{settings} = {};
    for my $parse_fv (split /\s*,\s*/,$parse_val) {
      my ($parse_fvf, $parse_fvv) = split /\s*=>\s*/,$parse_fv;
      $parse_fpd->{$parse_urldef}{settings}{$parse_fvf} = $parse_fvv;
    }
  }

  # Check whether or not we need to fill in the form. We will look at the URL ($url).
  # Process form if applicable 
  if (exists $parse_fpd->{$url}) {
    # Get any cookie received with the form
    my @rsc = ();
    $res->scan(sub {
      my $hdr = shift;
      my $val = shift;
      if ($hdr =~ /Set-Cookie/i) {
        my @cva = split /[\s]*;[\s]*/, $val;
        for my $cv (@cva) {
          if ($cv !~ /^domain=/i && $cv !~ /^path=/i && $cv !~ /^expires=/i) {
            push @rsc, $cv;
          }
        }
      }
    });

    # We start to process form
    $r->log->error("Processing form at $url");
    my $tree = HTML::TreeBuilder->new_from_content($res->content);
    my @forms = $tree->find_by_tag_name('form');
    my $g = $forms[0];
    for my $f (@forms) {
      if ($f->attr('name') eq $parse_fpd->{$url}{formName}) {
        $g = $f;
        last;
      }
    }
    my $thef = HTTP::Request::Form->new($g, $url);
    $r->log->error("Sending form data to ".$thef->link);
    for my $f (keys %{$parse_fpd->{$url}{settings}}) {
      my $fv = eval $parse_fpd->{$url}{settings}{$f};
      # FIXME: "eval" is dangerous...
      if (!defined $fv) { $fv = $parse_fpd->{$url}{settings}{$f}; }
      $thef->field($f, $fv);
    }
    my $freq = $thef->press();
    # Send along with the new request the cookies received from the client
    # in the original request and those received from the remote server in the
    # initial response. Only a big "Cookie" line is sent containing all valid
    # cookies (FIXME: it could overflow)
    my $fcv = "";
    my $fcv_field = 'Cookie';
    for my $hdr (keys %hdrs) {
      if ($hdr =~ /^Cookie$/i) {
        $fcv_field = $hdr;
        for my $ckv (split /[\s]*;[\s]*/,$hdrs{$hdr}) {
          my ($cn, $cv) = split /[\s]*=[\s]*/,$ckv,2;
          if ($cn !~ /$TMECookie/i) {
            if ($r->dir_config('Cookie_AddString')) {
              # Remove suffix string
              my $suffix = $r->dir_config('Cookie_AddString');
              $cn =~ s/_$suffix$//; 
            }
            if ($fcv ne "") { $fcv .= "; $cn=$cv"; }
            else { $fcv = "$cn=$cv"; }
          }
        }
      }
    }
    for my $cv (@rsc) {
      if ($fcv ne "") { $fcv .= "; $cv"; }
      else { $fcv = "$cv"; }
    }
    $freq->header ($old_LWP ? $fcv_field : ':'.$fcv_field, $fcv) if ($fcv ne "");

    # Finally, we send the new request with filled-in form
    $res = $ua->request($freq);
  }

  $r->status($res->code);

  # Process and tidy received headers:
  # Cookies are transformed so they can be sent back to the originating server
  # Redirects are adapted to the rewriten URLs
  $r->log->notice("***** Parsing response (code ".$res->code.") *****") if $debug;
  $res->scan(sub {
    my $hdr = shift;
    my $val = shift;
    $r->log->notice("Hdr: $hdr // Val: $val") if $debug;
    if ($hdr =~ /Set-Cookie/i) { 
      my @cva = split /[\s]*;[\s]*/, $val;
      my $cfv = undef;
      for my $cv (@cva) {
        if ($cv !~ /^domain=/i) {

          if ( $r->dir_config('Cookie_AddString') && ($cv !~ /^path=/i) && ($cv !~ /^expires=/i) ) {
            # Add suffix string
            my $suffix = $r->dir_config('Cookie_AddString');
            my ($cookie_name, $cookie_value) = split /[\s]*=[\s]*/,$cv,2;
            $cookie_name .= '_' . $suffix;
            $cv = $cookie_name . '=' . $cookie_value;
            $r->log->notice("Added suffix in Set-Cookie: cn=$cookie_name // cv=$cookie_value") if $debug;
          }

          if (!defined $cfv){ $cfv = $cv; }
          else { $cfv .= "; $cv"; }
        }
      }
      $val = $cfv;
      $r->cgi_header_out ($hdr, $val);
    }
    elsif ($hdr =~ /Location/i) {
      $val = &fixLink($r,$val,$url);
      $r->header_out ($hdr, $val);
    }
    else { $r->header_out ($hdr, $val); }
  });
  my $outString = "";

  # Fix empty Content-type header
  if (!$res->content_type && $res->code==200) {
    $res->content_type('text/html');
    $r->log->notice('Fixed empty/missing Content-type header (set to \'' . $res->content_type . '\')') if $debug;
  }

  $r->log->notice('Content-Type: ' . $res->content_type) if $debug;

  # Parse and rewrite HTML content
  my $print_parsed_response = 0;
  if ( ($res->content_type =~ /html/i) and !($r->dir_config('Raw_Redirect_All')) ) {
    $print_parsed_response = 1;
    $r->log->notice("Rewriting HTML content") if $debug;
    my $content = $res->content;
    my $baseHref = "";
    my $inScript = 0;
    my $p = HTML::TokeParser->new(\$content)
            || $r->log->error("No content available from remote URL $url: $!"); 
    while (my $tolkens = $p->get_token ) {
      if ($tolkens->[0] eq 'T') {
        if ($inScript || ($r->dir_config('Redirect_All') && $r->dir_config('Redirect_All') eq "1")) {
          $tolkens->[1] = &rewriteString($r, 'PAPI_Redirect', $tolkens->[1]);
        }
        $outString .= $tolkens->[1];
      }
      elsif ($tolkens->[0] eq 'E') {
      # End script rewriting
        if ($tolkens->[1] =~ /^script$/i ||
            $tolkens->[1] =~ /^style$/i) { $inScript = 0; }
        $outString .= "</" . $tolkens->[1] . ">";
      }
      elsif ($tolkens->[0] eq 'S') {
      # End script rewriting. It will be activated if starting a script tag
        $inScript = 0;
        my $text = $tolkens->[4];
        my @oldLink = ();
        my @newLink = ();
        if ($tolkens->[1] =~ /^a$/i) {
          if ($tolkens->[2]{href}) {
            push @newLink, &fixAnchor($r,$tolkens->[2]{href},$url);

            push @oldLink, regexEscape($tolkens->[2]{href});
          }
          if ($tolkens->[2]{onclick}) {
            push @newLink, &fixAnchor($r,$tolkens->[2]{onclick},$url);
            push @oldLink, regexEscape($tolkens->[2]{onclick});
          }
        }
        elsif ($tolkens->[1] =~ /^base$/i) {
          if ($tolkens->[2]{href}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{href},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{href});
            $url = $tolkens->[2]{href};
          }
        }
        elsif ($tolkens->[1] =~ /^meta$/i) {
          if ($tolkens->[2]{content} =~ /url/i) {
            my ($junk,$tmpLink) = split (/=/, $tolkens->[2]{content});
            $newLink[0] = &fixLink($r,$tmpLink,$url);
            $oldLink[0] = regexEscape($tmpLink);
          }
        }
        elsif ($tolkens->[1] =~ /^(area|link)$/i) {
          if ($tolkens->[2]{href}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{href},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{href}); 
          }
        }
        elsif ($tolkens->[1] =~ /^(frame|img|iframe)$/i) {
          if ($tolkens->[2]{src}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{src},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{src});
          }
        }
        elsif ($tolkens->[1] =~ /^form$/i) {
          if ($tolkens->[2]{action}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{action},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{action});
          }
        }
        elsif ($tolkens->[1] =~ /^(td|body)$/i) {
          if ($tolkens->[2]{background}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{background},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{background});
          }
        }
        # Activate rewriting for the content of style specs (imported URLs...)
        elsif ($tolkens->[1] =~ /^style$/i) { $inScript = 1; }
        elsif ($tolkens->[1] =~ /^script$/i) {
          if ($tolkens->[2]{src}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{src},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{src});
          }
        # Inline script (not from a source object). Activate rewriting
          else { $inScript = 1; }
        }
        elsif ($tolkens->[1] =~ /^option$/i) {
          if ($tolkens->[2]{value}) {
            $newLink[0] = &fixLink($r,$tolkens->[2]{value},$url);
            $oldLink[0] = regexEscape($tolkens->[2]{value});
          }
        }
        elsif ($tolkens->[1] =~ /^input$/i) {
          if ($tolkens->[2]{value}) {
            push @newLink, &fixAnchor($r,$tolkens->[2]{value},$url);
            push @oldLink, regexEscape($tolkens->[2]{value});
          }
          if ($tolkens->[2]{src}) {
            push @newLink, &fixAnchor($r,$tolkens->[2]{src},$url);
            push @oldLink, regexEscape($tolkens->[2]{src});
          }
        }
        # Some servers send character entities inside URLs, that make the regex
        # substiting $oldLink by $newLink fail. Up to now, only '&amp;' -> '&'
        # is considered
        if ($#newLink >= 0) {
          $text =~ s(\&amp;)(&)gs;
        }
        
        for my $i (0..$#newLink) {
          $text =~ s($oldLink[$i])($newLink[$i])gs;
        }
        
        $outString .= $text;

      } else { $outString .= $tolkens->[4]; }
    }
  }
 
  # Non-HTML content or Raw_Redirect_All. Check for MIME types and URL patterns
  # where rewriting is applicable
  else {
    $outString = $res->content;
    my $rw = 0;

    my $mime_regexp = $r->dir_config('Rewrite_MIME_Types');
    if ($mime_regexp) {
      if ( !( $r->dir_config('Rewrite_MIME_Types_Extended') && $r->dir_config('Rewrite_MIME_Types_Extended') eq "1" ) ) {
        # Not in extended mode: we should escape "Rewrite_MIME_Types" parameter
        $mime_regexp = escape_string($mime_regexp);
      }

      # Rewritable MIME-Types are separated by whitespace
      $mime_regexp =~ s/(\s)+/\|/g;
    }

    if ($mime_regexp and $res->content_type =~ /$mime_regexp/i) {
 	$rw = 1;
	$r->log->notice("Rewriting \'" . $res->content_type . "\' content according to Rewrite_MIME_Types") if $debug;

    } elsif ( $res->content_type =~ /html/i and $r->dir_config('Raw_Redirect_All') ) {
      $rw = 1;
      $r->log->notice("Rewriting HTML content (Raw_Redirect_All)") if $debug;

    } else {
      if ($r->dir_config('Rewrite_URL_Pattern_Extended') && $r->dir_config('Rewrite_URL_Pattern_Extended') eq "1" ) {
        # Extended mode
        for my $patt ($r->dir_config->get('Rewrite_URL_Pattern')) {
          my $rpt = eval $patt;
          if (!defined $rpt) { $rpt = $patt; }
          if ($url =~ /$rpt/) { $rw = 1; $r->log->notice("(Extended) Rewriting \'" . $res->content_type . "\' content according to Rewrite_URL_Pattern. Rule: $rpt") if $debug; last; }
        }
      } else {
        # Normal mode
        for my $patt ($r->dir_config->get('Rewrite_URL_Pattern')) {
          my $escaped_patt = escape_string($patt);
          if ($url =~ /$escaped_patt/) { $rw = 1; $r->log->notice("(Normal) Rewriting \'" . $res->content_type . "\' content according to Rewrite_URL_Pattern. Rule: $patt") if $debug; last; }
        } 
      }
    }

    if ($rw) {
      my $server_name = $r->get_server_name;            # proxy's server name
      my $base_name = $r->location();                   # Normally '/'  (i.e. "<Location />" in httpd.conf)
      my $parsed_uri=Apache::URI->parse($r);
      my $scheme = $parsed_uri->scheme;                 # 'http' or 'https'
      $print_parsed_response = 1;

      # FIXME:  Esto solo funciona para Remote_URL. No funcionara correctamente para Remote_Domain.

      if ($r->dir_config('Strip_Location') && $r->dir_config('Strip_Location') eq "1") {
        $base_name =~ s%/+$%%g; 
      } else {
        $base_name = "";
      }

      if ( ($scheme eq 'http' and $r->server->port != 80) or ($scheme eq 'https' and $r->server->port != 443) ) { $server_name .= ":".$r->server->port; }
      my $proxyurl = $scheme . "://" . $server_name . $base_name;

      # In normal case (Strip_Location = 0), only the server and port will be translated. E.g:
      # - $host_dest = 'http://1.2.3.4:5000'  (always without trailing '/')
      # - $proxy_url = 'http://proxy:port'
      #
      # When Strip_Location = 1, we will have something like:
      # - $host_dest = 'http://1.2.3.4:5000'   (always without trailing '/')
      # - $proxy_url = 'http://proxy:port/location'  (always without trailing '/')
      #
      # Caveat: it may create conflicts with URLs or strings like: 'http://1.2.3.4:50000' (note the additional '0').
      # Anyway, it's a not so common case and still we could fix it by using PAPI_Redirect, as in:
      # PerlAddVar PAPI_Redirect "http://proxy:port0 http://1.2.3.4:50000" ==> This would revert a previous translation 
      # PerlAddVar PAPI_Redirect "http://proxy:port/location0 http://1.2.3.4:50000" ==> Idem when Strip_Location = 1
      #
      # Why don't we include trailing '/' in $host_dest? Well, we could have a legal URL 'http://1.2.3.4:5000' and it wouldn't
      # be translated!

      $outString =~ s/$host_dest/$proxyurl/ig;
      $outString = &rewriteString($r, 'PAPI_Redirect', $outString);
    }
  }
  $r->header_out('Content-Length'=> length($outString));
  $r->content_type($res->content_type);
  $r->send_http_header;
  $r->print($outString);
  if ($print_parsed_response && $debug && $debug == 2) {
    $r->log->notice("Parsed response:");
    $r->log->notice("=======================================================================");
    for my $line (split(/\n/, $outString)) {
      $r->log->notice($line);
    }
    $r->log->notice("=======================================================================");
  }
  return(OK);
}



##################################################################
# sub fixAnchor
#
# This is a wrapper to fixLink for dealing with JavaScript anchors
#
##################################################################
sub fixAnchor {
  my $r = shift;
  my $link = shift;
  my $baseLink = shift;

  $r->log->notice("Entering sub fixAnchor [ link=$link, baseLink=$baseLink ]") if $debug;

  my $rtv = &fixLink($r, $link, $baseLink);
  # If a special method is in use, apply redirects
  # At the moment, only "javascript:" is considered
  if ($rtv =~ /^javascript:/i) {
    $rtv = &rewriteString($r, 'PAPI_Redirect', $rtv);
  }
  return $rtv;
}



###############################
# sub fixLink
#
# Parameters:
#
# $r - an Apache request object.
# $link - a reference to another resource.
#     It can be a URL or a relative reference.
#     fixLink will do the right thing with it.
# $baseLink - the link pointing to the current resource.
#
# Returns:
# a fixed URL using this module again.
#################################
sub fixLink {
  my $r = shift;
  my $link = shift;
  my $baseLink = shift;
  my $server_name = $r->get_server_name;
  my $i;
  my $urlPath = "";
  my $hostName = "";
  my $base_name=$r->location();
  # Location must end with '/'. Better to force it...
  $base_name =~ s%/+$%%g;
  $base_name .= '/';

  my $tdn = undef;
  my $savedlink = $link;

  $r->log->notice("Entering sub fixLink [ link=$link, baseLink=$baseLink ]") if $debug;

  # Guess "scheme" (RFC 2616) of reverse proxy url ('http' o 'https')
  my $parsed_uri=Apache::URI->parse($r);
  my $scheme = $parsed_uri->scheme;

  my ($protocol, $hostname, @dirs);
  ($protocol, $hostname) = split (/\/\//, $baseLink);
  ($hostname,@dirs) = split (/\//, $hostname);
  if ($dirs[1] =~ /\w/) { $urlPath = join ("/", @dirs); }
  else { $urlPath = $dirs[0] unless ($baseLink =~ /\w$/); }
  if ($baseLink =~ /\w$/ && $dirs[1] =~ /\w/) {
    for ($i = length($urlPath); $i > 0; $i--) {
      last if (substr($urlPath, $i, 1) eq "/");
    }
    $urlPath = substr ($urlPath, 0, $i) if ($i && $urlPath =~ /\//);
  }

  if ($r->dir_config('Remote_Domain')) {
    $tdn = $r->dir_config('Remote_Domain');
    my ($thn) = ($hostname=~m(^(.+?)\.($tdn)$));
    $base_name = "/".$thn.$base_name;
  }
  $base_name=~ s/$path_dest$//;
	
  $hostname = "$protocol//$hostname";
  if ( ($scheme eq 'http' and $r->server->port != 80) or ($scheme eq 'https' and $r->server->port != 443) ) { $server_name .= ":".$r->server->port; }
  if ($link =~ /^\//) {
    $link = $scheme . "://". $server_name. $base_name . $link;
  }
  elsif ($link =~ /^\.\./) {
    $link = $scheme . "://". $server_name.$base_name. "/".$urlPath. "/".$link;
  }
  elsif ($link =~ /^$host_dest/i) {
    $link =~ s/^$host_dest//i;
    $link = $scheme . "://" . $server_name . $base_name . $link;
  }
  elsif (defined $tdn && $link =~ /^\w+?\.$tdn/i) {
    my ($up1, $up2) = split (/\/\//, $link);
    my ($thn) = ($up2=~m(^(.+?)\.($tdn)));
    $link =~ s/^(.+?)\.$tdn//i;
    $link = $scheme . "://" . $server_name . "/" . $thn . $link;
  }
  $link = &rewriteString($r, 'PAPI_Redirect', $link);

  $r->log->notice("Rewrite: $savedlink  -->  $link") if ($debug and ($savedlink ne $link));
  return $link;
}



sub regexEscape {
  my $url = shift;
  # This silly little regex fixes -[]()*+?|&# in the URL for me.
  # withhout this regex, any of these characters in a URL will
  # cause a server error (unless they resolve into something
  # sensible to regex, in which case the server does something
  # magical and unpredictable with the URL)
  $url =~ s/(\-|\[|\]|\(|\)|\*|\+|\?|\||\&|\#)/\\$1/g;
  return $url;
}

1;

__END__

=head1 NAME

B<Apache::RewritingProxyTME> - mod_perl Apache handler to process requests for
remote resources. Or with other words... a full I<reverse proxy> implementation
:-)


=head1 SYNOPSIS

 sub handler {
     my $r = shift;
     
     ...

     $r->handler("perl-script");
     $r->set_handlers(PerlAccessHandler => undef);
     $r->set_handlers(PerlHandler => [Apache::RewritingProxyTME]);

     ...
     
 }

			      
=head1 DESCRIPTION

This module is hardly based on PAPI::RewritingProxy. It has been changed to
support HTTPS and can be configured using PerlSetVar and PerlAddVar statements
in httpd.conf, in a way that no other PAPI modules are necessary. There are
many other improvements, new features and fixes (see Changelog).

The B<handler> function in this package processes requests, redirecting them
according to the settings of either the B<Remote_URL>
or B<Remote_Domain> configuration parameter. The URI to be
requested is build according to the value of B<Strip_Location>. If the
parameter B<Proxy_Server> has been specified, the handler will use it to
satisfy the request. If HTTP Authentication is in use at any remote
server the B<HTTP_Auth>
configuration parameter can be employed to seamlessly access
those restricted resources. If there are B<Post_Redirect> directives they
are processed.

If the request (once translated into the remote server/domain) matches any of
the URLs identified for automatic form processing by the configuration directive
B<Form_Processor>, the handler will proceed to request the specified URL,
fill the defined values into the form and send it to the remote server. The
response sent back to the user will be that sent by the remote server in
response to the automatically filled form.

Once the response from the remote server is received, links are rewritten 
according to:

1. The value of B<Remote_URL> or B<Remote_Domain> (takes precedence over
B<PAPI_Redirect> directives).

2. The setting of the B<Rewrite_MIME_Types>
configuration parameter.

3. The rules defined by the B<Rewrite_URL_Pattern>
configuration parameters.

4. The rules defined by the B<PAPI_Redirect>
configuration parameters.

5. The setting of the B<Redirect_All> configuration parameter.

The following elements are processed by the rewriting procedures:

1. Contents of B<Location> HTTP headers (redirects).

2. The I<href> attribute of the B<a>, B<area>, B<base>, and B<link> HTML tags.

3. The I<content> attribute of the B<meta> HTML tag.

4. The I<src> attribute of the B<frame>, B<iframe>, B<img>, B<input>, and
B<script> HTML tags.

5. The I<action> attribute of the B<form> HTML tag.

6. The I<background> attribute of the B<body> and B<td> HTML tags.

7. The I<value> attribute of the B<input> and B<option> HTML tags.

8. The contents of any B<script> and B<style> HTML element.

9. JavaScript attributes of HTML tags.

The logic of the process is the following:

=over 4

=item 1.	If HTML content ("Content-Type" field contains "html" string):

=over 4

=item 1.1.	  If B<Raw_Redirect_All>, a "raw" rewrite will be performed. In other
words, rewrite is applied to the whole response contents.

=item 1.2.	  Otherwise:

=over 4

=item 1.2.1.	    If B<Redirect_All>, smart-rewrite will be performed (only applies
to HTML tags, active or not).

=item 1.2.2.	    Otherwise, smart-rewrite will be performed (only applies to active
HTML tags).

=back

=back

=item 2.	If non-HTML content:

=over 4

=item 2.1.	  If "Content-Type" matches B<Rewrite_MIME_Types>, a "raw" rewrite is
performed.

=item 2.2.	  If url (of final web) matches B<Rewrite_URL_Pattern>, a "raw"
rewrite will be performed.

=back

=back

=head1 CONFIGURATION

The following configuration directives are permitted:

=over 4

=item B<Remote_URL>

The URL of final web (it should be written without the final "/"). Proxy will
connect to it to retrieve information requested by the browser. Example:

	PerlSetVar Remote_URL http://10.253.65.16:81

=item B<Remote_Domain>

This is the alternative to B<Remote_URL>. It produces the following URL
translation:

  http://proxy/host/location/...  <--> http://host.remote.domain/location/.... 

Example:

	PerlSetVar Remote_Domain ieee.org


=item B<Remote_Domain_HTTPS> [1|0]

By default (value 0), B<Remote_Domain> will translate as follows:

  http://proxy/host/location/...   <--> http://host.remote.domain/location/.... 
  https://proxy/host/location/...  <--> http://host.remote.domain/location/.... 

If we activate this option (value 1) we force the use of HTTPS:

  http://proxy/host/location/...   <--> https://host.remote.domain/location/.... 
  https://proxy/host/location/...  <--> https://host.remote.domain/location/.... 

Example:

	PerlSetVar Remote_Domain_HTTPS 1

Note: type of URL (http or https) used to access the proxy (i.e. the source
of "translations") only depends on Apache configuration (implemented as the
reverse proxy).

=item B<Redirect_All> [1|0]

By default (value 0) rewriting rules are only applied to active links (for
instance, inside a HREF tag or SRC parameter in a IMG or SCRIPT tag). If you
want to be applied to a more extensible set of places, switch this option on.
Example:

	PerlSetVar Redirect_All 1

=item B<Raw_Redirect_All> [1|0]

By default (value 0) rewrite engine is smart enough to only parse HTML content
and will analyze certain fields/tags which could contain URLs (or not). In
some cases this analysis may fail or give unexpected results. Only for those
cases it is useful to switch this directive on, forcing the processing of the
whole page contents (bypassing the parser). Example:

	PerlSetVar Raw_Redirect_All 1

=item B<Strip_Location> [1|0]

By default (value 0), when the proxy receives a request it will pass URI
without being altered to final web. If you switch this option on, the proxy
will not use <Location> value (it will not be sent to final web). This feature
may be useful if you pretend to map several remote webs to different proxy
locations. Example:

  <Location /web>
    PerlSetVar Remote_URL                   http://one.dom.ain
    PerlSetVar Strip_Location               1
  </Location>
  <Location /anotherweb>
    PerlSetVar Remote_URL                   http://two.dom.ain
    PerlSetVar Strip_Location               1
  </Location>

Will be translated as:

  http://proxy/web/location/...	        <-->   http://one.dom.ain/location/... 
  http://proxy/anotherweb/location/...  <-->   http://two.dom.ain/location/...

=item B<Proxy_Server>

Intermediate proxy (in URL form) to be used to reach final web. If this proxy
requests for authentication, we should use also B<Proxy_User> and
B<Proxy_Pass> directives. Example:

	PerlSetVar Proxy_Server http://192.168.0.9:8080

=item B<Proxy_User>

If the intermediate proxy requests for authentication, we should specify here
the username. Example:

	PerlSetVar Proxy_User "username"

=item B<Proxy_Pass>

If the intermediate proxy requests for authentication, we should specify here
the password. Example:

	PerlSetVar Proxy_Pass "password"

=item B<HTTP_Auth>

Valid for automagically respond to any HTTP auth request ("Basic" and/or
"Digest") sent by final web. May be several directives of this kind (they are
distinguished by the "realm" field, specifying the remote resource). Example:

	PerlAddVar HTTP_Auth "realm::username::password"

=item B<Form_Processor>

Forms auto-fill in. Several directives of this kind may exist. Each one should
be include URL of the page containing the form as well as field-value pairs to
be automatically fill in by the module. Example:

	PerlAddVar Form_Processor "http://192.168.0.3/dir/login.php usuario =>
roman, pass => secret"

=item B<PAPI_Redirect_Extended> [1|0]

By default (value 0) it disables the use of regexp in B<PAPI_Redirect>
directives (it will avoid having to escape special chars). If we activate
"extended" mode (value 1) behaviour will be the common one of PAPI, I mean,
you will be able to use regexp, etc (but you will have to use "\" char to
escape). Example:

	PerlSetVar PAPI_Redirect_Extended 1

=item B<PAPI_Redirect_Delim>

Regexp that defines the delimitier used to separate lookup string from the
replace string, in B<PAPI_Redirect> directives. By default: '\s+' (i.e. one or
more "white" chars). Example:

	PerlSetVar PAPI_Redirect_Delim '\s+'

=item B<PAPI_Redirect>

Rewrite rule. In "extended" mode, it admits regexps and proxy variables (that
will be evaluated before performing the rewrite). Several B<PAPI_Redirect>
directives may exist, specified by PerlAddVar (instead of PerlSetVar).

Example of "normal" mode (default):

	PerlAddVar PAPI_Redirect "http://192.168.0.3 https://proxy:12638"

Example of "extended" mode:

	PerlAddVar PAPI_Redirect "([\w]+).ieee.org MPROXYNAME/$1/"

=item B<Post_Redirect_Extended> [1|0]

By default (value 0) it disables the use of regexp in B<Post_Redirect>
directives (it will avoid having to escape special chars). If we activate
"extended" mode (value 1) behaviour will be the common one of PAPI, I mean,
you will be able to use regexp, etc (but you will have to use "\" char to
escape). Example:

	PerlSetVar Post_Redirect_Extended 1

=item B<Post_Redirect_Delim>

Regexp that defines the delimitier used to separate lookup string from the
replace string, in B<PAPI_Redirect> directives. By default: '\s+' (i.e. one or
more "white" chars). Example:

	PerlSetVar Post_Redirect_Delim '\s+'

=item B<Post_Redirect>

Rewrite rule to be applied to POST requests before being sent to final web
server. Syntax is simlar to the one used by B<PAPI_Redirect>.

Example in "normal" mode (default):

	PerlAddVar Post_Redirect "cadena nuevacadena"

Example in "extended" mode:	

	PerlAddVar Post_Redirect "^(validaOper\#.=item\#.*) $1\n"

=item B<Rewrite_MIME_Types_Extended> [1|0]

By default (value 0) it disables the use of regexp in B<Rewrite_MIME_Types>
directives (it will avoid having to escape special chars). If we activate
"extended" mode (value 1) behaviour will be the common one of PAPI, I mean,
you will be able to use regexp, etc (but you will have to use "\" char to
escape). Example:

	PerlSetVar Rewrite_MIME_Types_Extended 1

=item B<Rewrite_MIME_Types>

By default the module only applies its rewrite engine to HTML content. If you
want also to be applied to additional content, it should be specified by using
this directive (enter the list of permitted MIME-Types separated by one space).
Example:

	PerlSetVar Rewrite_MIME_Types "application/x-javascript tipo/mime"

=item B<Rewrite_URL_Pattern_Extended> [1|0]

By default (value 0) it disables the use of regexp in B<Rewrite_URL_Pattern>
directives (it will avoid having to escape special chars). If we activate
"extended" mode (value 1) behaviour will be the common one of PAPI, I mean,
you will be able to use regexp, etc (but you will have to use "\" char to
escape). Example:

	PerlSetVar Rewrite_URL_Pattern_Extended 1

=item B<Rewrite_URL_Pattern>

B<PAPI_Redirect> rewrite rules affect HTML content and also specified MIME-
Types. This directive enables matching against "URL patterns". If the URL
(final web, not proxy) matches one or some of these (PerlAddVar) directives,
rewrite engine will act over this page. Regexp may be used (in "extended"
mode).

Example of "normal" mode (default):

	PerlAddVar Rewrite_URL_Pattern "http://www.tsm.es/file.js"

Example of "extended" mode:

	PerlAddVar Rewrite_URL_Pattern "\.js$"

=item B<Cookie_AddString>

String (usually the vhost port) to be added to the name of all cookies (except
Apache cookie-auth) (separated by "_" char). This modification is only visible
in browser-proxy stretch (it is undone before reaching proxy-final web
stretch). In this way, cookies conflict between different applications using
the same (or similar) nomenclature is transparently avoided. Example:

	PerlSetVar Cookie_AddString 10713

=item B<NoCache> [1|0]

If enabled (value 1), reverse proxy will add "Pragma: no-cache" and
"Cache-Control: no-cache" headers to HTTP requests sent to final web server,
asking for avoiding caching. Useful for debugging purposes. Example:

	PerlSetVar NoCache 1

=item B<Debug> [0|1|2]

By default (value 0), the module only logs a few events (e.g. the URL being
accessed). If enabled (value 1 or 2), a more verbose debugging will be
produced. The most detailed logging is produced with a value of 2 (the whole
content of requests and responses will be printed). Example:

	PerlSetVar Debug 1

=back

=head1 AUTHOR

Roman Medina-Heigl Hernandez (Telefonica Moviles) - Spain.

Original PAPI module was written by:

Rodrigo Castro (RedIRIS) - Spain.

Diego R. Lopez (RedIRIS) - Spain.

=cut
